#include "LightingApplication.h"

//The compression and simulation settings:
#define NUM_CLUSTERS   1
#define NUM_PCAVECTORS 24
#define ORDER          6

struct MODEL_VERTEX
{
	float x, y, z;
	float nx, ny, nz;
	float ClusterID;
	float PCAData[24];
};

//The declaration now includes several additional "slots" to hold the SH data.
D3DVERTEXELEMENT9 Declaration[] = 
{
    {0,  0,  D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,    0},
    {0,  12, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,      0},
    {0,  24, D3DDECLTYPE_FLOAT1, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 0},
    {0,  28, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 1},
    {0,  44, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 2},
    {0,  60, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 3},
    {0,  76, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 4},
    {0,  92, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 5},
    {0, 108, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT, 6},
    D3DDECL_END()
};

//A little utility function
void SetColorValue(D3DCOLORVALUE *pV, float a, float r, float g, float b) {pV->a = a; pV->r = r; pV->g = g; pV->b = b;};
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }

CLightingApplication::CLightingApplication()
{
	long  Count     = 0;
	m_VertexShaders[Count].pFileName     = "c:\\Shaders\\SH_VS.hsl";
	m_VertexShaders[Count].pName         = "SH";
	m_VertexShaders[Count++].pTargetName = "vs_1_1";

	//This sample only really uses one mesh...
	m_Models[0].pFileName = "c:\\test.x";
	m_pSHDataFileName = "c:\\testresults.dat";
	
	//Set up the declaration and the stride to match what the application needs
	m_pVertexElements = Declaration;
	m_Stride          = sizeof(MODEL_VERTEX);
 
	//Clear some of the basic SH data arrays
	m_ClusterBases = NULL;
	m_ClusteredPCA = NULL;

    m_pPRTBuffer     = NULL;
    m_pPRTCompBuffer = NULL;
}

CLightingApplication::~CLightingApplication()
{
}

BOOL CLightingApplication::Initialize()
{		
	if (m_pSHDataFileName)
		LoadSHDataFromFile(m_pSHDataFileName);

    float       Radius;
    D3DXVECTOR3 Center;
	MODEL_VERTEX *pVertices = NULL;
    m_Models[0].pModel->LockVertexBuffer(0, (void**)&pVertices);
	D3DXComputeBoundingSphere((D3DXVECTOR3 *)pVertices, m_Models[0].NumVertices, m_Stride, &Center, &Radius);
	D3DXVECTOR3 EyePosition = Center + D3DXVECTOR3(-Radius, Radius, 0.0f);
	//Build a view matrix using the eye position
	D3DXMatrixLookAtLH(&m_View, &(Center + D3DXVECTOR3(-Radius * 1.0f, Radius * 1.0f, 0.0f)), &Center, &D3DXVECTOR3(0.0f, 1.0f, 0.0f));
	m_Models[0].pModel->UnlockVertexBuffer();

	return TRUE;
}

void CLightingApplication::Render()
{
	//Set the directional light vector. 
	D3DXVECTOR4 LightVector(sin((float)GetTickCount() / 3000.0f), fabs(cos((float)GetTickCount() / 3000.0f)), 0.0f, 0.0f);
	//D3DXVECTOR4 LightVector(0.0f, 1.0f, 0.0f, 0.0f);
	D3DXVec4Transform(&LightVector, &LightVector, &m_InverseWorld);
	D3DXVec4Normalize(&LightVector, &LightVector);
	m_pD3DDevice->SetVertexShaderConstantF(4, (float *)&LightVector, 1);

	//Set the light color
	D3DXVECTOR4 LightColor(1.0f, 1.0f, 1.0f, 1.0f);
	m_pD3DDevice->SetVertexShaderConstantF(6, (float *)&LightColor, 1);

	//If you want, add code to make sure that the SH material color and
	//this color are the same.
	D3DXVECTOR4 DiffuseColor(1.0f, 1.0f, 1.0f, 0.0f);
	m_pD3DDevice->SetVertexShaderConstantF(7, (float *)&DiffuseColor, 1);
	
	//Provide the concatenated transformation
	m_ShaderMatrix = m_World * m_View * m_Projection;
	m_pD3DDevice->SetVertexShaderConstantF(0, (float *)&m_ShaderMatrix, 4);

	m_pD3DDevice->SetVertexDeclaration(m_pVertexDeclaration);
	m_pD3DDevice->SetStreamSource(0, m_Models[m_CurrentModel].pModelVertices, 0, m_Stride);
	m_pD3DDevice->SetIndices(m_Models[m_CurrentModel].pModelIndices);

	SetSHShaderConstants((D3DXVECTOR3 *)&LightVector);

	m_pD3DDevice->SetVertexShader(m_VertexShaders[m_CurrentVertexShader].pShader);

	//These calls do the actual drawing
	m_pD3DDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0,
		                               m_Models[m_CurrentModel].NumVertices,
									   0, m_Models[m_CurrentModel].NumFaces);
}

BOOL CLightingApplication::Terminate()
{
    if (m_pPRTBuffer) 
	{
		m_pPRTBuffer->Release();
		m_pPRTBuffer = NULL;
	}
    if (m_pPRTCompBuffer) 
	{
		m_pPRTCompBuffer->Release();
		m_pPRTCompBuffer = NULL;
	}
	SAFE_DELETE_ARRAY(m_ClusterBases);
	SAFE_DELETE_ARRAY(m_ClusteredPCA);
	return TRUE;
}

BOOL CLightingApplication::HandleMessage(MSG *pMessage)
{	
	return CSampleApplication::HandleMessage(pMessage);
}

//Load precached SH data from a file. This is a very slimmed down version of the 
//SH vertex shader code in the SDK.
HRESULT CLightingApplication::LoadSHDataFromFile(char *pFileName)
{
   if (m_pPRTBuffer) 
	{
		m_pPRTBuffer->Release();
		m_pPRTBuffer = NULL;
	}
    if (m_pPRTCompBuffer) 
	{
		m_pPRTCompBuffer->Release();
		m_pPRTCompBuffer = NULL;
	}

	HRESULT h = D3DXLoadPRTBufferFromFile(pFileName, &m_pPRTBuffer);
    h = D3DXCreatePRTCompBuffer(D3DXSHCQUAL_FASTLOWQUALITY, NUM_CLUSTERS,
							NUM_PCAVECTORS, m_pPRTBuffer, &m_pPRTCompBuffer);

    h = m_pPRTCompBuffer->NormalizeData();

    UINT dwNumSamples  = m_pPRTCompBuffer->GetNumSamples();
    UINT dwNumCoeffs   = m_pPRTCompBuffer->GetNumCoeffs();
    UINT dwNumChannels = m_pPRTCompBuffer->GetNumChannels();
    UINT dwNumClusters = m_pPRTCompBuffer->GetNumClusters();
    UINT dwNumPCA      = m_pPRTCompBuffer->GetNumPCA();

    UINT* pClusterIDs = new UINT[dwNumSamples];
    m_pPRTCompBuffer->ExtractClusterIDs(pClusterIDs);
	MODEL_VERTEX *pV = NULL;
    m_Models[0].pModel->LockVertexBuffer(0, (void**)&pV);
    for(UINT CurrentVertex = 0; CurrentVertex < dwNumSamples; CurrentVertex++) 
    {
		//Generate the cluster Index as a float using the ID from the array and the cluster size.
		float CurrentClusterIndex = (float)(pClusterIDs[CurrentVertex] * (1 + 3 * (NUM_PCAVECTORS / 4))); 
        pV[CurrentVertex].ClusterID = CurrentClusterIndex;
    }
    m_Models[0].pModel->UnlockVertexBuffer();
    SAFE_DELETE_ARRAY(pClusterIDs);

    h = m_pPRTCompBuffer->ExtractToMesh(dwNumPCA, D3DDECLUSAGE_BLENDWEIGHT, 1, m_Models[0].pModel);

	long SizeOfCluster = ORDER * ORDER * 3 * (NUM_PCAVECTORS + 1);
	m_ClusterBases = new float[SizeOfCluster * NUM_CLUSTERS];

    for(long CurrentCluster = 0; CurrentCluster < NUM_CLUSTERS; CurrentCluster++) 
        m_pPRTCompBuffer->ExtractBasis( CurrentCluster, &m_ClusterBases[CurrentCluster * SizeOfCluster]);

    m_ClusteredPCA = new float[NUM_CLUSTERS * (4 + 3 * NUM_PCAVECTORS)];

    return S_OK;
}

//Define this to use the cone light
//#define USE_CONE_LIGHT

HRESULT CLightingApplication::SetSHShaderConstants(D3DXVECTOR3 *pObjectSpaceLightVector)
{
    float LightValues[3][ORDER*ORDER];  

    //Project the light vector into SH.
#ifdef USE_CONE_LIGHT
	float Radius = (0.51f + 0.5f * sin((float)GetTickCount() / 1000.0f)) * 3.0f;
	D3DXSHEvalConeLight(ORDER, &(D3DXVECTOR3(0.0f, 1.0f, 0.0f)),Radius, 1.0f, 1.0f, 1.0f,
						LightValues[0], LightValues[1], LightValues[2]);
#else
	D3DXSHEvalDirectionalLight(ORDER, pObjectSpaceLightVector, 1.0f, 1.0f, 1.0f,
                               LightValues[0], LightValues[1], LightValues[2]);
#endif


	for (long Cluster = 0; Cluster < NUM_CLUSTERS; Cluster++) 
    {
 		//Find the dot product of the light solution and the mean of each cluster.
		long PCAIndex  = Cluster * (4 + 3 * NUM_PCAVECTORS);
		long BasisIndex = Cluster * ORDER * ORDER * 3 * (NUM_PCAVECTORS + 1);

        m_ClusteredPCA[PCAIndex + 0] = D3DXSHDot(ORDER, &m_ClusterBases[BasisIndex + 0 * ORDER * ORDER], LightValues[0]);
        m_ClusteredPCA[PCAIndex + 1] = D3DXSHDot(ORDER, &m_ClusterBases[BasisIndex + 1 * ORDER * ORDER], LightValues[1]);
        m_ClusteredPCA[PCAIndex + 2] = D3DXSHDot(ORDER, &m_ClusterBases[BasisIndex + 2 * ORDER * ORDER], LightValues[2]);
        m_ClusteredPCA[PCAIndex + 3] = 0.0f;

        //Find the dot product of the light solution and the individual bases of each cluster
        float *pPCAVector = &m_ClusteredPCA[Cluster * (4 + 3 * NUM_PCAVECTORS) + 4];
        for(long CurrentPCA = 0; CurrentPCA < NUM_PCAVECTORS; CurrentPCA++) 
        {
            long Offset = BasisIndex + (CurrentPCA + 1) * ORDER * ORDER * 3;

            pPCAVector[0 * NUM_PCAVECTORS + CurrentPCA] = D3DXSHDot(ORDER, &m_ClusterBases[Offset + 0 * ORDER * ORDER], LightValues[0]);
            pPCAVector[1 * NUM_PCAVECTORS + CurrentPCA] = D3DXSHDot(ORDER, &m_ClusterBases[Offset + 1 * ORDER * ORDER], LightValues[1]);
            pPCAVector[2 * NUM_PCAVECTORS + CurrentPCA] = D3DXSHDot(ORDER, &m_ClusterBases[Offset + 2 * ORDER * ORDER], LightValues[2]);
        }
    }

	m_pD3DDevice->SetVertexShaderConstantF(4, m_ClusteredPCA, (NUM_CLUSTERS * (4 + 3 * NUM_PCAVECTORS)) / 4);

    return S_OK;
}

