#include "LightingApplication.h"

static CLightingApplication *g_pLightingApplication = NULL; 

LRESULT WINAPI EntryMessageHandler(HWND hWnd, UINT Message, 
							  WPARAM wParam, LPARAM lParam)
{
	return g_pLightingApplication->MessageHandler(hWnd, Message, wParam, lParam);
}

INT WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT)
{
	g_pLightingApplication = new CLightingApplication();
    g_pLightingApplication->Go();
	delete (g_pLightingApplication);
	return 0;
}


CSampleApplication::CSampleApplication()
{
	memset(m_VertexShaders, 0, NUM_VERTEX_SHADERS * sizeof(VERTEX_SHADER_INFO));
	memset(m_PixelShaders, 0,  NUM_PIXEL_SHADERS  * sizeof(PIXEL_SHADER_INFO));
	memset(m_Models, 0, NUM_MODELS * sizeof(MODEL_INFO));

	m_CurrentVertexShader = 0;
	m_CurrentPixelShader  = 0;
	m_CurrentModel        = 0;
	m_LastChangeTime      = 0;
}

void CSampleApplication::Go()
{
	m_WindowWidth  = 640;
	m_WindowHeight = 480;

	WNDCLASSEX WindowClass = {sizeof(WNDCLASSEX), 
							  CS_CLASSDC,
							  EntryMessageHandler,
							  0,
							  0, 
							  GetModuleHandle(NULL),
							  NULL,
							  NULL,
							  NULL,
							  NULL,
							  "Sample Application",
							  NULL};

	RegisterClassEx(&WindowClass);

    m_hWnd = CreateWindow("Sample Application",
							"Sample Application", 
							WS_OVERLAPPEDWINDOW,
							0,
							0,
							m_WindowWidth,
							m_WindowHeight,
							GetDesktopWindow(),
							NULL,
							WindowClass.hInstance,
							NULL);

	if(m_hWnd)
		ShowWindow(m_hWnd, SW_SHOW);

	m_Continue = TRUE;

	InitializeD3D();

	D3DCAPS9 Caps;
	m_pD3D->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &Caps);
	if (Caps.VertexShaderVersion == D3DVS_VERSION(1,1))
	{
		if (FAILED(CreateWindowed(m_hWnd, D3DDEVTYPE_HAL, D3DCREATE_HARDWARE_VERTEXPROCESSING)))
			return;
	}
	else
	{
		if (FAILED(CreateWindowed(m_hWnd, D3DDEVTYPE_HAL, D3DCREATE_SOFTWARE_VERTEXPROCESSING)))
			return;
	}

	m_pD3DDevice->CreateVertexDeclaration(m_pVertexElements, &m_pVertexDeclaration);

	CreateShaders();
	CreateVertices();

	if (!Initialize())
		return;

	//Do some basic matrix setup.
	m_FOV         = D3DX_PI / 4.0f;
	m_AspectRatio = (float)m_WindowWidth / (float)m_WindowHeight;
	D3DXMatrixPerspectiveFovLH(&m_Projection, m_FOV, m_AspectRatio, 1.0f, 10000.0f);

	m_WorldAngle = 0.0f;
	D3DXMatrixRotationY(&m_World, m_WorldAngle);
	D3DXMatrixInverse(&m_InverseWorld, NULL, &m_World);

    D3DXCreateFont(m_pD3DDevice, 20, 10, 0, FALSE, 0, 0, 0, 0, 0, "Arial", &m_pFont);
	m_TextRect.left  = m_TextRect.top = 0;
	m_TextRect.right  = 640;
	m_TextRect.bottom = 480;

	QueryPerformanceFrequency((LARGE_INTEGER *)&m_HighResFrequency);
	__int64 StartTime;
	__int64 EndTime;
	
	MSG Message; 
	PeekMessage(&Message, 0, 0, 0, PM_REMOVE);

	while (Message.message != WM_QUIT && m_Continue)
    {
		QueryPerformanceCounter((LARGE_INTEGER *)&StartTime);
		
		PreRender();
		Render();

		QueryPerformanceCounter((LARGE_INTEGER *)&EndTime);
		sprintf(m_TextBuffer, "%3.2f fps", (float)(m_HighResFrequency / (double)(EndTime - StartTime)));
		m_pFont->DrawText(NULL, m_TextBuffer, -1, &m_TextRect, DT_LEFT, 0xffffffff);

		if (D3DERR_DEVICELOST == PostRender())
		{
			PreReset();

			RestoreDevice();

			PostReset();
		}


		TranslateMessage(&Message);
        DispatchMessage(&Message);
		PeekMessage(&Message, 0, 0, 0, PM_REMOVE);

		m_Continue = HandleMessage(&Message);
    }

	Terminate();

	m_pFont->Release();

	DestroyDevice();
}

void CSampleApplication::PreRender()
{
	m_pD3DDevice->Clear(0, NULL,
						D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
						D3DCOLOR_XRGB(100, 100, 255), 1.0f, 0);

	m_pD3DDevice->BeginScene();

	//Provide some rudimentary data to the user about the current shader
	sprintf(m_TextBuffer, "\n%s", m_VertexShaders[m_CurrentVertexShader].pName);
	m_pFont->DrawText(NULL, m_TextBuffer, -1, &m_TextRect, DT_LEFT, 0xffffffff);
			
	return;
}

HRESULT CSampleApplication::PostRender()
{
	m_pD3DDevice->EndScene();
	return m_pD3DDevice->Present(NULL, NULL, NULL, NULL);
}


CSampleApplication::~CSampleApplication()
{
	//Basically, just make sure everything is properly released.
	if (m_pVertexDeclaration)
		m_pVertexDeclaration->Release();

	for (long i = 0; i < NUM_MODELS; i++)
	{
		if (m_Models[i].pModelVertices)
			m_Models[i].pModelVertices->Release();

		if (m_Models[i].pModelIndices)
			m_Models[i].pModelIndices->Release();

		if (m_Models[i].pModel)
			m_Models[i].pModel->Release();
	}

	for (i = 0; i < NUM_VERTEX_SHADERS; i++)
	{
		if (m_VertexShaders[i].pShader)
			m_VertexShaders[i].pShader->Release();
	}

	for (i = 0; i < NUM_PIXEL_SHADERS; i++)
	{
		if (m_PixelShaders[i].pShader)
			m_PixelShaders[i].pShader->Release();
	}

	if (m_pD3D)
		m_pD3D->Release();

	m_pD3D = NULL;
}

LRESULT WINAPI CSampleApplication::MessageHandler(HWND hWnd, 
												UINT Message, 
												WPARAM wParam,
												LPARAM lParam)
{
	switch(Message)
    {
		case WM_DESTROY:
            PostQuitMessage(0);
            return 0;
    }
	return DefWindowProc(hWnd, Message, wParam, lParam);
}

HRESULT CSampleApplication::InitializeD3D()
{
	m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	m_pD3DDevice = NULL;
	return S_OK;
}

HRESULT CSampleApplication::CreateDevice(
			D3DDEVICE_CREATION_PARAMETERS *pCreateParms,
			D3DPRESENT_PARAMETERS         *pPresentParms)
{
	memcpy(&m_CreationParameters, pCreateParms, 
		   sizeof(D3DDEVICE_CREATION_PARAMETERS));
	memcpy(&m_PresentParameters, pPresentParms,
		   sizeof(D3DPRESENT_PARAMETERS));

	return m_pD3D->CreateDevice(pCreateParms->AdapterOrdinal, 
		                        pCreateParms->DeviceType,
						        pCreateParms->hFocusWindow,
						        pCreateParms->BehaviorFlags,
						        pPresentParms,
						        &m_pD3DDevice);
}

HRESULT CSampleApplication::DestroyDevice()
{
	if (m_pD3DDevice)
		m_pD3DDevice->Release();

	m_pD3DDevice = NULL;

	return S_OK;
}


HRESULT CSampleApplication::RestoreDevice()
{
	HRESULT Result = m_pD3DDevice->TestCooperativeLevel();

	while(Result == D3DERR_DEVICELOST)
	{
		while(Result != D3DERR_DEVICENOTRESET)
		{
			Sleep(1000);

			MSG Message;
			PeekMessage(&Message, 0, 0, 0, PM_REMOVE);
			TranslateMessage(&Message);
			DispatchMessage(&Message);

			Result = m_pD3DDevice->TestCooperativeLevel();
		}

		if (FAILED(m_pD3DDevice->Reset(&m_PresentParameters)))
			Result = D3DERR_DEVICELOST;
	}	

	return S_OK;
}

HRESULT CSampleApplication::CreateWindowed(HWND WindowHandle, 
							             D3DDEVTYPE DeviceType, 
							             DWORD Behavior)
{
	D3DDISPLAYMODE CurrentMode;
	if (SUCCEEDED(m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,
								                &CurrentMode)))
	{
		ZeroMemory(&m_PresentParameters, 
				   sizeof(D3DPRESENT_PARAMETERS));
		m_PresentParameters.Windowed = TRUE;
		m_PresentParameters.SwapEffect = D3DSWAPEFFECT_DISCARD;
		m_PresentParameters.BackBufferFormat = CurrentMode.Format;
		m_PresentParameters.EnableAutoDepthStencil = TRUE;
		m_PresentParameters.AutoDepthStencilFormat = D3DFMT_D24S8;

		m_CreationParameters.AdapterOrdinal = 
											D3DADAPTER_DEFAULT;
		m_CreationParameters.DeviceType     = DeviceType;
		m_CreationParameters.hFocusWindow   = WindowHandle;
		m_CreationParameters.BehaviorFlags  = Behavior;

		return CreateDevice(&m_CreationParameters, 
							&m_PresentParameters);
	}

	return E_FAIL;
}

BOOL CSampleApplication::HandleMessage(MSG *pMessage)
{
	if (pMessage->message == WM_KEYDOWN && pMessage->wParam == VK_ESCAPE)
		return FALSE;

	//These keys provide basic user interaction functionality.
	if (pMessage->message == WM_KEYDOWN)
	{
		if (pMessage->wParam == VK_F1)
		{
			if (!m_VertexShaders[++m_CurrentVertexShader].pName)
				m_CurrentVertexShader = 0;
		}

		if (pMessage->wParam == VK_F2 && GetTickCount() - m_LastChangeTime > CHANGE_DELAY)
		{
			m_LastChangeTime = GetTickCount();

			if (!m_Models[++m_CurrentModel].pName)
				m_CurrentModel = 0;
		}

		if (pMessage->wParam == VK_F3)
		{
			if (!m_PixelShaders[++m_CurrentPixelShader].pName)
				m_CurrentPixelShader = 0;
		}

		//The movement increments are tuned to work well on most machines, but
		//they might require tweaking.
		if (pMessage->wParam == VK_UP)
		{
			m_FOV = max(0.0f, m_FOV - 0.01f);
			D3DXMatrixPerspectiveFovLH(&m_Projection, m_FOV, m_AspectRatio, 1.0f, 10000.0f);
		}

		if (pMessage->wParam == VK_DOWN)
		{
			m_FOV = min(D3DX_PI / 2.0f, m_FOV + 0.01f);
			D3DXMatrixPerspectiveFovLH(&m_Projection, m_FOV, m_AspectRatio, 1.0f, 10000.0f);
		}

		if (pMessage->wParam == VK_LEFT)
		{
			m_WorldAngle += 0.05f;
			D3DXMatrixRotationY(&m_World, m_WorldAngle);
		}

		if (pMessage->wParam == VK_RIGHT)
		{
			m_WorldAngle -= 0.05f;
			D3DXMatrixRotationY(&m_World, m_WorldAngle);
		}

		//Assume that some matrix changes were made and reset everything
		D3DXMatrixInverse(&m_InverseWorld, NULL, &m_World);
	}

	return TRUE;
}

void CSampleApplication::CreateShaders()
{	
	LPD3DXBUFFER pShaderCode = NULL;

	for (long i = 0; i < NUM_VERTEX_SHADERS; i++)
	{
		if (m_VertexShaders[i].pFileName)
		{
			D3DXCompileShaderFromFile(m_VertexShaders[i].pFileName, NULL, NULL, "main", m_VertexShaders[i].pTargetName, NULL, &pShaderCode, NULL, NULL);
			m_pD3DDevice->CreateVertexShader((DWORD*)pShaderCode->GetBufferPointer(), &(m_VertexShaders[i].pShader));
			pShaderCode->Release();
		}
	}
}

void CSampleApplication::CreateVertices()
{
	//This code loads one mesh and gets all the information about
	//the mesh to streamline the rendering process later.
	for (long i = 0; i < NUM_MODELS; i++)
	{
		if (m_Models[i].pFileName)
		{		
			D3DXLoadMeshFromX(m_Models[i].pFileName, D3DXMESH_MANAGED,
							  m_pD3DDevice, NULL, NULL, NULL, NULL, &m_Models[i].pModel);
		}
		else
			break; //Create some shapes...
	}

	//The tail end of the model array will be a bunch of D3DX shapes
	m_Models[i].pName = "Box";
	D3DXCreateBox(m_pD3DDevice, 10.0f, 10.0f, 10.0f, &m_Models[i].pModel, NULL);
	m_Models[++i].pName = "Cylinder";
	D3DXCreateCylinder(m_pD3DDevice, 5.0f, 5.0f, 10.0f, 40, 40, &m_Models[i].pModel, NULL);
	m_Models[++i].pName = "Torus";
	D3DXCreateTorus(m_pD3DDevice, 2.0f, 5.0f, 40, 40, &m_Models[i].pModel, NULL);
	m_Models[++i].pName = "Sphere";
	D3DXCreateSphere(m_pD3DDevice, 5.0f, 40, 40, &m_Models[i].pModel, NULL);
	m_Models[++i].pName = "Teapot";
	D3DXCreateTeapot(m_pD3DDevice, &m_Models[i].pModel, NULL);

	//In addition to loading the mesh, it also makes sure that the vertices are
	//in the proper format and that the correct normal vectors exist.
	for (; i >= 0; i--)
	{
		LPD3DXMESH pTempMesh;

		m_Models[i].pModel->CloneMesh(D3DXMESH_MANAGED, m_pVertexElements, m_pD3DDevice, &pTempMesh);
		m_Models[i].pModel->Release();
		m_Models[i].pModel = pTempMesh;

 			DWORD *rgdwAdjacency = NULL;
			rgdwAdjacency = new DWORD[m_Models[i].pModel->GetNumFaces() * 3];
			m_Models[i].pModel->ConvertPointRepsToAdjacency(NULL, rgdwAdjacency);
			m_Models[i].pModel->OptimizeInplace(D3DXMESHOPT_VERTEXCACHE, rgdwAdjacency, NULL, NULL, NULL);
			delete []rgdwAdjacency;

		m_Models[i].pModel->GetVertexBuffer(&m_Models[i].pModelVertices);
		m_Models[i].pModel->GetIndexBuffer(&m_Models[i].pModelIndices);
		
		//D3DXComputeNormals(m_Models[i].pModel, NULL);

		m_Models[i].NumFaces    = m_Models[i].pModel->GetNumFaces();
		m_Models[i].NumVertices = m_Models[i].pModel->GetNumVertices();
 	}
}

void CSampleApplication::Render(){}
BOOL CSampleApplication::Initialize(){return TRUE;}
BOOL CSampleApplication::Terminate(){return TRUE;}
BOOL CSampleApplication::PreReset(){return TRUE;}
BOOL CSampleApplication::PostReset(){return TRUE;}



